# -*- coding: utf-8 -*-
import pytest

from titan.cli import cli_orbit


def cli_orbit_argv(capsys, argv):
    cli_orbit(argv.split())
    out, err = capsys.readouterr()
    assert err == ''
    return out.rstrip()

def test_cli_orbit_date_equinox(capsys):
    assert cli_orbit_argv(capsys, '1980-02-22') == '0.004273686299484325'

def test_cli_orbit_date_equinox_verbose(capsys):
    assert cli_orbit_argv(capsys, '-v 1980-02-22') == u'1980-02-22 -> Ls: 0.00º'


def test_cli_orbit_Ls_equinox(capsys):
    assert cli_orbit_argv(capsys, '0') == '2009-07-30'

def test_cli_orbit_Ls_equinox_verbose(capsys):
    assert cli_orbit_argv(capsys, '-v 0') == u'Ls: 0º (+ 1 Titan year since 1980)-> 2009-07-30'

def test_cli_orbit_Ls_equinox_offset(capsys):
    assert cli_orbit_argv(capsys, '-o 0 0') == '1980-02-22'

def test_cli_orbit_Ls_equinox_offset_verbose(capsys):
    assert cli_orbit_argv(capsys, '-v -o 0 0') == u'Ls: 0º (+ 0 Titan year since 1980)-> 1980-02-22'


def test_cli_orbit_date_multiple(capsys):
    assert cli_orbit_argv(capsys, '2009-07-30 2017-05-14') == '0.004273686299484325\n90.35962529291561'

def test_cli_orbit_Ls_multiple(capsys):
    assert cli_orbit_argv(capsys, '0 10 20 30') == '2009-07-30\n2010-05-21\n2011-03-18\n2012-01-18'


def test_cli_orbit_date_fmt_slash(capsys):
    assert cli_orbit_argv(capsys, '1980/02/22') == '0.004273686299484325'

def test_cli_orbit_date_fmt_datetime(capsys):
    assert cli_orbit_argv(capsys, '1980-02-22T01:02:03') == '0.004273686299484325'


def test_cli_orbit_date_err(capsys):
    with pytest.raises(ValueError):
        cli_orbit_argv(capsys, 'foo')



# -*- coding: utf-8 -*-
import numpy as np

# Constantes
UA = 149597870.7  # 1 Astronomical unit (km)

# Default orbital parameters
ORBIT = {
    'saturn_orbit': 15.945,             # 1 Saturn orbit (days)
    'sun_orbit':    10751,              # 1 Sun orbit (days)
    'obliquity':    26.730882944988142,
    'vernal_equinox': '1980-02-22',     # Date of the first vernal equinox (before Voyager 1)
    'ellipse': {
        'A': 6.1664830805512354,
        'B': 6.0482745790986066,
        'C': 101.03535416292833,
    },
}

def readDate(date):
    '''Read date as datetime64'''
    if type(date) == str:
        date = np.datetime64(
            date.replace('/', '-').replace(' ', 'T').split('T')[0], 'D')
    return date

class Orbit:
    '''Titan orbit functions and parametes'''

    def __init__(self):
        '''Init default parameters'''
        self.Tday = ORBIT['saturn_orbit']

        # Default orbit parameters calculated with NAIF Space Kernels
        self.obl   = ORBIT['obliquity']
        self.orbit = np.timedelta64(ORBIT['sun_orbit'], 'D')
        self.eq_v  = np.datetime64(ORBIT['vernal_equinox'])
        self.A     = ORBIT['ellipse']['A']
        self.B     = ORBIT['ellipse']['B']
        self.C     = ORBIT['ellipse']['C']
        return

    def __repr__(self):
        return 'Titan orbit functions and parametes'

    def Ls(self, date, eps=1.e-7, imax=25):
        '''Calculate the solar longitude corresponding to a date.

        Parameters
        -----------
        date : str, numpy.datetime64
            Input date (YYYY-MM-DD or YYYY/MM/DD or YYYY-MM-DDThh:mm:ss.ms)
        eps : float, optional
            Precision of the convergence
        imax : int, optional
            Number maximum of iteration to reach the convergence, throw a ValueError otherwise.

        Note
        -----
        The value of Ls is the solution of a transcendental equation which
        is numerically solved with the Newton method:

        L_s^0 = 360 · (Date - Eq^V)/Orbit) - B
        
        L_s^(n+1) = L_s^n - (
            L_s^n - L_s^0 + A · sin(2·pi·(L_s^n - C)/360)
        )/(
            1 + A · 2·pi/360 · cos(2·pi·(L_s^n - C)/360)
        )

        Return
        -------
        Ls : real
            Solar latitude corresponding to the input date
        '''
        date = readDate(date)

        Ls_0 = ( (360.*(date - self.eq_v).astype(int))/self.orbit.astype(float) - self.B ) % 360
        Ls   = Ls_0
        for ii in range(imax):
            dLs = - (Ls - Ls_0 + self.A * np.sin(2*np.pi*(Ls - self.C)/360.))         \
                / (1 + self.A * 2*np.pi/360. * np.sin(2*np.pi*(Ls - self.C)/360.))
            Ls = Ls + dLs
            if np.abs(dLs) < eps:
                break
        else:
            raise ValueError('Max number of iteration reach without getting convergence.')
        return Ls % 360

    def date(self, Ls, Ty=0):
        '''Calculate the date corresponding to a solar longitude.

        Parameters
        -----------
        Ls : real
            Input solar latitude
        Ty : int, optional
            Number of Titan year after 1980-02-22 (Vernal Equinox before Voyager 1)

        Return
        -------
        date : numpy.datetime64
            Date corresponding to the input solar latitude
        '''
        date = np.round(
            self.orbit.astype(int)/360. * (
                Ls + self.A * np.sin(2*np.pi*(Ls - self.C)/360.) + self.B + 360 * Ty
            )
        )
        return self.eq_v + np.timedelta64(int(date), 'D')


orbit = Orbit()



# -*- coding: utf-8 -*-
import argparse

from ..orbit import orbit

def cli_orbit(argv=None):
    parser = argparse.ArgumentParser(
        description="""Get Titan (Saturn's moon) orbital constrains""")

    parser.add_argument('values', nargs='+', type=str, metavar='date|Ls',
                        help='Calendar date  (YYYY-MM-DD or YYYY/MM/DD or YYYY-MM-DDThh:mm:ss.ms) or Solar longitude value(s)')

    parser.add_argument('-o', type=int, default=1, metavar='offset',
                        help='Titan year offset since 1980 (default: 1)')

    parser.add_argument('--verbose', '-v', action='store_true', help='Verbose output')

    args, _ = parser.parse_known_args(argv)
    for val in args.values:
        try:
            Ls = float(val)
            date = orbit.date(Ls, Ty=args.o)
            
            if args.verbose:
                print("Ls: {}º (+ {} Titan year since 1980)-> {}".format(val, args.o, date))
            else:
                print(date)

        except ValueError:
            Ls = orbit.Ls(val)

            if args.verbose:
                print("{} -> Ls: {:.2f}º".format(val, Ls))
            else:
                print(Ls)



# -*- coding: utf-8 -*-
import pytest

from titan import orbit

EQ_V = '1980-02-22'

def test_Ls_eq_v():
    Ls = orbit.Ls(EQ_V)
    assert pytest.approx(Ls, abs=0.05) == 0

def test_date_eq_v():
    date = orbit.date(0)
    assert str(date) == EQ_V

def test_Ls_convergence_err():
    with pytest.raises(ValueError):
        orbit.Ls(EQ_V, imax=3)